#!/bin/bash
#
####### WARNING WARNING WARNING WARNING WARNING WARNING WARNING ######
######															######
######	Run this script on directories that are quiescent.		######
######		Changing the names of open files/directories		######
######		could result in data loss.							######
######															######
######	I recommend first running this script in Preview Mode	######
######		to see what will be changed.						######
######															######
######	Changing the names of directories and files will affect	######
######		the functionality of certain apps. For example,		######
######		Microsoft Office for Mac uses a tilde to designate	######
######		temporary recovery files and directories, removing 	######
######		them will prevent normal recovery.					######
######															######
####### WARNING WARNING WARNING WARNING WARNING WARNING WARNING ######
#
#
#
######################################################################
#
#	fix_unsafe_filenames.sh
#
#	Samuel Herschbein
#	sam@SamIsIT.com
#	sherschbein@me.com
#
#	Version:
#		1.0.0 - 2020-03-17 Initial release
#
#	I wrote this to fix directory and file names on a Synology NAS that is accessed from macOS 10.6-10.14 plus
#		Windows 7, 8.1, and 10. Many files on the NAS originated in older versions of Windows, macOS, Mac OS X, and Classic Mac.
#
#	The NAS is littered with decade+ old Microsoft temp files of the forms ~XXXXXXX and XXXXXX~X where X =[A-Z0-9]. The script will optionally
#		delete these files. In Preview mode the script only says they would be deleted.
#
#	The NAS also has many metadata/resource fork orphans (files prefixed with ._ that don't have the matching original file without the ._).
#		 In Preview mode the script only says they would be deleted.
#
#	Coded in macOS Sierra 10.12 and Mojave 10.14 with GNU bash version 3.2.57(1).
#		macOS 10.11 through 10.14 all have this same version of bash.
#	Heavily tested on a High Sierra Mac mini with a Carbon Copy Cloner clone of a Synology NAS.
#
#	For all directories in the hierarchy (when not in Preview Mode):
#		For all directory and file names:
#			Remove leading dashes.
#			Decode \o, \oo, and \ooo octal values, removing unprintable characters.
#			Decode URL encoded %xx values, removing unprintable characters.
#			Change characters with special meanings to different OSes to underscores.
#		Optionally delete old Microsoft temp files.
#		Optionally delete orphaned metadata/resource fork files.
#
#	This script accepts 0 to 4 options.
#		For brief help the first option must be -? (other options are ignored).
#		The order of the remaining options does not matter.
#		For Preview Mode pass -p.
#		To delete (or display in Preview Mode) Microsoft temp files pass -m.
#		To delete (or display in Preview Mode) orphaned metadata/resource fork files pass -o.
#		An optional path can be passed, if it is absent the current/working directory is used.
#
#	Modifying the booted macOS's directories could make macOS unstable or unbootable. To prevent this only
#		directories in /Users and /Volumes are allowed. There's no protection for a boot volume mounted in /Volumes.
#
#	This script assumes you have permission to execute the commands find, mv, rm, and cd in the directory hierarchy.
#		If you're unsure, run the script in Preview Mode and redirect stdout and stderr as show below.
#
#	Preview Mode traverses the hierarchy and reports what would be changed (nothing is modified).
#	An effective way to use Preview Mode is to redirect both stdout and stderr to a single file and then peruse it
#		using an editor like BBEdit. To redirect stdout & stderr into the same file:
#		./fix_unsafe_filenames.sh -p -m -o &> fix_stdout_stderr.txt
#
#	The characters changed are octal values \o \oo \ooo, URL encoded %xx values, and the characters `~*-\|:"<>/
#	Octal values are in the form \o, \oo, and \ooo where o=0..7. ASCII values are 0..177.
#		The script discards unprintable ASCIIs (000..037 & 177) as well as 200..777.
#		There can be ambiguities, for example \777. Is this:
#			\7 (unprintable bell) followed by "77", or
#			\77 ("?") followed by "7", or
#			\777 some two-byte character?
#		For simplicity, the script treats 0..3 octal digits following \ as an octals, it discards octals > 177.
# 	URL encoded %xx values are three characters of the form %xx where x is a hexadecimal value. For example, %2F decodes to / (forward slash).
#		Unprintables are discarded:  %00..%1F & %7F..%FF. The rest are converted to their ASCII character.
#	The individual characters changed are, from the top left of a US keyboard moving right then down, with unshifted first:
#		` acute/grave acccent: used in Unix/Linux commands.
#		~ tilde: used in Unix/Linux commands.
#		* asterisk/star: reserved in Unix/Linux, used by all OSes as a wildcard.
#		- dash: only leading dashes are removed to avoid confusion with Unix/Linux command options.
#		\ backslash: reserved in DOS/Win.
#		| pipe/vertical bar: reserved in DOS/Win, used in Unix/Linux commands.
#		: colon: reserved in DOS/Win and macOS.
#		" double-quote: reserved in DOS/Win.
#		< left angle bracket: used in Unix/Linux commands.
#		> right angle bracket: used in Unix/Linux commands.
#		/ slash: reserved in Unix/Linux, allowed in macOS. At the Unix file level macOS converts displayed slashes to colons so they're not
#				treated as directory delimiters. Some badly-written macOS apps don't convert slashes to colons.
#
#	A changed file/directory name could be identical to an existing file/directory name. A duplicate name
#		without an extension will be renamed "My Item 2", while a duplicate name with an extension will be renamed "My Item 2.ext".
#	The appended number starts at 2 and increments until a unique filename is found. Even a 32 bit system will have
#		4,294,967,294 positive appended numbers. That should be enough. ;)
#
#	This script can optionally delete orphaned metadata/resource fork files.
#	Metadata/resource fork files have the name of their matching companion (data fork) file prefixed with ._
#	For example:
#		foo.txt  	= the "original" or "companion" data fork file - normally visible to all users.
#		._foo.txt	= the metadata/resource fork file - normally hidden from macOS users since it starts with a period.
#	On HFS+ volumes the metadata/resource fork is handled transparently by both macOS and Unix commands.
#	A metadata/resource fork file is created when copying/saving to non-Mac volumes, e.g. SMB accessed volumes.
#	When a metadata file's companion file doesn't exist or is not in the same directory, it is considered an orphan.
#	An orphan is created when its companion file is deleted/moved by Windows, badly written Mac apps, buggy code, et al.
#
#	This script can optionally delete old Microsoft temp files.
#	Since these files have a tilde in their name they would normally be renamed. Since they are temps and very old deleting them
#		will usually not be an issue.
#	These files were 8 characters in length, consisting of capital letters and digits, with either the 1st or 7th character a tilde.
#	Examples: ~ABCD123 ABCD12~3.
#
#	The script ignores special macOS directories and custom icon files.
#
#	Some comments about this code:
#
#	The commands cd, mv, stat, and printf have -- added to them as the first option so directories with leading dashes/hyphens are accepted.
#		Some printf commands do not need the -- so they don't have it.
#
#	I'm aware that bash functions and recursion are relatively slow because new sub-shells are spawned.
#		Using a recursive function greatly simplified the algorithm that traverses the hierarchy.
#
#	To keep the script succinct bash's one-line if-then and if-then-else else statements are used:
#		[[ test ]] && if_True || if_False
#	I find the more concise script easier to debug in spite of the more obtuse construct. YMMV.
#
######################################################################
#
#
######################################################################
#
#	Preferences
#
######################################################################
#
#												#	These are used when displaying a file/directory's dates.
												#	They must be bash standard format as used by sed, et al.
dateCreatedFormat="+%Y-%m-%d %I:%M:%S %p"
dateModifiedFormat="+%A, %Y-%m-%d %I:%M:%S %p"
#
itemTypeForDirectory="Folder"					#	I prefer Folder instead of Directory since I share results with non-geeks.
titleForDirectory="folder"
#
												#	These can get long if more descriptive.
itemTypeForOldMicrosoftTempFile="Old Microsoft temp file"
blurbForOldMicrosoftTempFiles="old Microsoft temp files"
itemTypeForOrphanedMetadataFile="Orphaned metadata file"
blurbForOrphanedMetadataFiles="orphaned metadata files"
#
#
######################################################################
#
#	Globals
#
######################################################################
#
customIconFilename="$(printf "Icon\r")"			#	The custom icon filename is Icon\r, it displays as Icon?. The printf puts the actual unprintable into customIconFilename.
#
#
######################################################################
#
#	Recursively_Fix_File_and_Folder_Names
#
#	Traverses the working directory checking both file and folder names.
#		Also checks for old Microsoft temp files and orphaned metadata/resource fork files.
#
#	If in Preview Mode, display what would be changed, otherwise change it.
#
######################################################################
Recursively_Fix_File_and_Folder_Names () {
	local firstChangedItem=true
	local myWD="$(pwd)"
	local itemIsOrphanedMetadataFile
																						#	The find has maxdepth 1 since sub-directories are searched recursively.
																						#	Do the find on files first and then directories.
																						#	For legibility use find's -s flag to search in lexicographical order.
																						#	Results are prepended with ./, cut that.
	local findResult=$(find -s . -maxdepth 1 -type f | cut -c 3-)$'\n'$(find -s . -maxdepth 1 -type d | cut -c 3-)
	printf "%s\n" "$findResult" | \
		while read -r itemName ; do
			if [[ "$itemName" != "" ]]; then
																						#	Never process macOS special directories, special filenames, or custom icon files.
				if ! { [[ -d "$itemName" ]] && [[ \
						"$itemName" == ".config" || \
						"$itemName" == ".DocumentRevisions-V100" || \
						"$itemName" == ".fseventsd" || \
						"$itemName" == ".local" || \
						"$itemName" == ".PKInstallSandboxManager-SystemSoftware" || \
						"$itemName" == ".Spotlight-V100" || \
						"$itemName" == ".TemporaryItems" || \
						"$itemName" == ".Trash" || \
						"$itemName" == ".Trashes" || \
						"$itemName" == ".vol" \
						]]; } && \
						! { [[ -f "$itemName" ]] && [[ \
						"$itemName" == ".DS_Store" || \
						"$itemName" == ".localized" || \
						"$itemName" == "$customIconFilename" \
						]]; }; then
					itemIsOldMicrosoftTempFile=false									#	To simplify the code these flags are set regardless if the item is a file or directory.
					itemIsOrphanedMetadataFile=false
					if [[ -d "$itemName" ]]; then
						itemIsDirectory=true											#	Item is a directory.
						itemType="$itemTypeForDirectory"
						duplicateFindQualifier="d"
					else
						itemIsDirectory=false											#	Item must be a file since find only searched for files and directories.
						itemType="File"
						duplicateFindQualifier="f"
						if [[ "${itemName:0:2}" == "._" && ! -f "${itemName:2:${#itemName}}" ]]; then
							itemIsOrphanedMetadataFile=true								#	Item is an orphaned ._ metadata/resource fork file.
							itemType="$itemTypeForOrphanedMetadataFile"
						elif [[ "${#itemName}" -eq 8 ]] &&  { [[ "${itemName:0:1}" == "~" ]] || [[ "${itemName:6:1}" == "~" ]]; }; then
							itemSansTilde="${itemName//\~/}"
							echo "$itemSansTilde"
							itemSansUppersDigits="${itemSansTilde##[A-Z0-9][A-Z0-9][A-Z0-9][A-Z0-9][A-Z0-9][A-Z0-9][A-Z0-9]}"
							[[ "$itemSansUppersDigits" == "" ]] && itemIsOldMicrosoftTempFile=true	#	Item is a Microsoft temp file.
						fi
					fi
																						#	Check the item's name first.
					newItemName="${itemName#"${itemName%%[!-]*}"}"						#	Remove all leading dashes/hyphens.
																						#	first remove octal triplets > 177, then use perl to convert backslashed octals to characters,
																						#		then use tr to discard unprintables (000..037 & 177).
					newItemName="$(printf "%s" "${newItemName//\\[2-7][0..7][0..7]/}" | perl -pe 's/\\([0-7]{1,3})/chr oct $1/eg' | tr -dc '[:print:]')"
					newItemName="${newItemName//%[0-18-9A-F][[:xdigit:]]/}"				#	Remove %xx URL encodings below space and above ASCII printables (00..1F & 80..FF).
																						#	perl converts remaining %xx to printable characters.
					newItemName="$(printf "%s" "$newItemName" | perl -pe 's/%([[:xdigit:]][[:xdigit:]])/chr hex $1/eg')"
					newItemName="${newItemName//[\`\~\*\|:\"\<\>\/]/\\}"				#	Substitute a backslash for all other unsafe characters.
					newItemName=$(printf "%s" "$newItemName" | tr -s "\\")				#	Replace multiple sequential backslashes with a single backslash.
					newItemName="${newItemName//\\/_}"									#	Replace backslashes with underscores.
																						#	newItemName is now either identical to itemName or a safe version of itemName.
																						#
					if $itemIsOldMicrosoftTempFile || $itemIsOrphanedMetadataFile; then	#	Handle old temp Microsoft files and orphans separately.
						if $firstChangedItem; then										#	If this item is the 1st changed/displayed in the current directory, display directory's name.
							printf -- "%s\n" "In $titleForDirectory \"${myWD##$path2cut}\":"
							firstChangedItem=false
						fi
						$itemIsOldMicrosoftTempFile && itemTypeTempOrphan="$itemTypeForOldMicrosoftTempFile" || itemTypeTempOrphan="$itemTypeForOrphanedMetadataFile"
						printf -- "\t%s" "$itemTypeTempOrphan \"${itemName//://}\" "
						if $previewMode && { { $itemIsOldMicrosoftTempFile && $deleteMicrosoftTempFiles; } || { $itemIsOrphanedMetadataFile && $deleteOrphanedMetadataFiles; }; }; then
							printf -- "%s\n" "would be deleted."
						elif ! $previewMode && { { $itemIsOldMicrosoftTempFile && $deleteMicrosoftTempFiles; } || { $itemIsOrphanedMetadataFile && $deleteOrphanedMetadataFiles; }; }; then
							rm -- "$itemName"											#	Delete the temp or orphan file.
							printf -- "%s\n" "was deleted."
						else
							printf -- "%s\n" "found."									#	The other 2 permutations of the if test variables, just display it.
						fi
					fi
																						#
																						#	Don't process an orphan or Microsoft temp name change if the item was, or would be, deleted.
					if [[ "$newItemName" != "$itemName" ]] && ! { $itemIsOldMicrosoftTempFile && $deleteMicrosoftTempFiles; } && ! { $itemIsOrphanedMetadataFile && $deleteOrphanedMetadataFiles; }; then
						if [ "-$duplicateFindQualifier" "$newItemName" ]; then			#	Single brackets used here so -$duplicateFindQualifier evaluated.
																						#	If new name exists, change new until non-existant name found.
							newItemNameB4Ext="${newItemName%.*}"
							if [[ "$newItemNameB4Ext" != "$newItemName" ]]; then
								hasExtension=true
								newItemNameExt=".${newItemName##*.}"					#	Get the extension (rightmost . and everything to its right).
							else
								hasExtension=false
							fi
							newItemNameSuffixNum=1
							while [ "-$duplicateFindQualifier" "$newItemName" ]; do		#	Single brackets used here so -$duplicateFindQualifier evaluated.
								((newItemNameSuffixNum++))
								newItemName="$newItemNameB4Ext $newItemNameSuffixNum"	#	Shortcut: if there's no extension newItemNameB4Ext=newItemName.
								[[ $hasExtension ]] && newItemName+="$newItemNameExt";	#	If there's an extension, add it to the new name.
							done
						fi
						if $firstChangedItem; then										#	If this item is the 1st changed in the current directory, display directory's name.
							printf -- "%s\n" "In $titleForDirectory \"${myWD##$path2cut}\":"
							firstChangedItem=false
						fi
						if $previewMode; then
							printf -- "\t%s\n" "$itemType \"${itemName//://}\" would be renamed to \"$newItemName\","
						else
							mv -- "$itemName" "$newItemName"							#	Rename the item.
							printf -- "\t%s\n" "$itemType \"${itemName//://}\" renamed \"$newItemName\","
							itemName="$newItemName";									#	Change itemName for commands below
						fi
						finderInfo="$(stat -s -- "$itemName")"							#	display creation & modification dates.
						createDate=$(date -j -f %s $(echo -n "$finderInfo" | sed -e 's/.*st_birthtime=\(.*\)st_blksize.*/\1/') "$dateCreatedFormat")
						modifyDate=$(date -j -f %s $(echo -n "$finderInfo" | sed -e 's/.*st_mtime=\(.*\)st_ctime.*/\1/') "$dateModifiedFormat")
						printf -- "\t\t%s\n" "Created: $createDate, last modified: $modifyDate"
					fi
																						#
					if $itemIsDirectory; then
						cdError="$(cd -- "$itemName" 2>&1)"								#	cd to directory (in sub-shell) with stderr redirected to stdout and assigned to variable.
						if [[ "$?" -ne 0 ]]; then										#	The most common error will be permission denied.
							printf -- "\n%s\n" "${cdError##*: } error cd'ing into \"${myWD##$path2cut}/$itemName\", this"
							printf -- "\t%s\n" "directory and its hierarchy will not be processed."
						else
							cd -- "$itemName"											#	Since above cd was in a sub-shell, cd in this shell.
							Recursively_Fix_File_and_Folder_Names						#	Traverse the child's hierarchy.
							cd ..														#	Move back to the parent directory before processing next item.
						fi
					fi
				fi
			fi
		done
}
#
#
######################################################################
#
#	Main Code
#
######################################################################
usageBlurb="\nUsage: fix_unsafe_filenames.sh [-? | -p | -m | -o | path ]\n"
usageBlurb+="\t[ -? ] Display this usage summary.\n"
usageBlurb+="\t[ -p ] Preview mode, display what would change without changing anything.\n"
usageBlurb+="\t[ -m ] Delete old Microsoft temp files of the format ABCDEF~G.\n"
usageBlurb+="\t[ -o ] Delete invisible orphaned metadata files (a.k.a. resource forks).\n"
usageBlurb+="\t[ path ] If path is empty, the current directory is used.\n"
usageBlurb+="\tTo prevent macOS corruption, only directories in /Users/* and /Volumes/* are allowed.\n\n"
																	#
																	#	See if user asked for help.
[[ "$1" == "-?" ]] && { printf "$usageBlurb"; exit; }
																	#	The option count must be 0..4.
[[ "$#" -gt 4 ]] && { printf "\nSpecify from 0 to 4 options.\n$usageBlurb"; exit; }
																	#
																	#	Default variables set by options
previewMode=false
deleteOrphanedMetadataFiles=false
deleteMicrosoftTempFiles=false
workingDIR=""
#
for option in ${1+"$@"}; do
	if [[ "$option" == "-p" ]]; then
		previewMode=true
	elif [[ "$option" == "-m" ]]; then
		deleteMicrosoftTempFiles=true
	elif [[ "$option" == "-o" ]]; then
		deleteOrphanedMetadataFiles=true
	elif [[ "${option:0:1}" == "-" ]] && ! [[ -d "$option" ]]; then	#	Has a leading - but isn't a valid option nor an existing directory.
			printf "\nUnrecognized option (or non-existant directory): \"$option\"\n$usageBlurb"
			exit
	elif [[ "$option" != "" ]]; then
		if [[ ! -d "$option" ]]; then								#	Has no leading dash, check that directory exists.
			printf -- "\n%s\n" "Directory does not exist: \"$option\""
			printf "$usageBlurb"
			exit
		else
			workingDIR="$option"
		fi
	fi
done
#
if [[ "$workingDIR" != "" ]]; then
	cdError="$(cd -- "$workingDIR" 2>&1)"							#	cd to directory (in sub-shell) with stderr redirected to stdout and assigned to variable.
	if [[ "$?" -ne 0 ]]; then										#	Since the directory exists (tested above), the most common error will be permission denied.
		printf -- "\n%s\n" "${cdError##*: } error cd'ing into \"$workingDIR\""
		printf "$usageBlurb"
		exit
	else
		cd -- "$workingDIR"											#	Since above cd was in a sub-shell, cd in this shell.
	fi
fi
																	#	Check for directories we should not be modifying.
if [[ ! "$(pwd)" =~ "/Users" ]] && [[ ! "$(pwd)" =~ "/Volumes" ]]; then
	printf -- "\n%s\n" "Directory path not allowed: \"$workingDIR\""
	printf "$usageBlurb"
	exit
fi
																	#
																	#	Finally, everything is OK.
																	#
startTime=$(date +%s)												#	Get the start time in seconds, let user know it in normal format.
printf "\nfix_unsafe_filenames started: $(date -r $startTime +"%Y-%m-%d %I:%M:%S %p")\n"
																	#
path2cut="$(dirname $(pwd))/"										#	Get pwd and strip off everything but the actual directory.

if $previewMode; then
	printf "\nPreview Mode: displaying unsafe names"
	$deleteMicrosoftTempFiles && printf " + $blurbForOldMicrosoftTempFiles"
	$deleteOrphanedMetadataFiles && printf " + $blurbForOrphanedMetadataFiles"
else
	printf "\nFixing unsafe names"
	if $deleteMicrosoftTempFiles; then
		printf " + deleting"
	else
		printf " + displaying"
	fi
	printf " $blurbForOldMicrosoftTempFiles"
	if $deleteOrphanedMetadataFiles; then
		printf " + deleting"
	else
		printf " + displaying"
	fi
	printf " $blurbForOrphanedMetadataFiles"
fi
printf -- " in:\n\t$(pwd)\n"
																	#
																	#
Recursively_Fix_File_and_Folder_Names								#	This function does all the work.
																	#
																	#
endTime="$(date +%s)"													#	Get the end time in seconds, calculate & display elapsed time.
printf "\nfix_unsafe_filenames ended: $(date -r "$endTime" +"%Y-%m-%d %I:%M:%S %p")\n"
elapsedTime="$(($endTime-$startTime))"
elapsedDays="$(($elapsedTime/86400))"
elapsedHours="$(($elapsedTime/3600%24))"
elapsedMinutes="$(($elapsedTime/60%60))"
elapsedSeconds="$(($elapsedTime%60))"
printf "Elapsed time: "
[[ "$elapsedDays" -gt 0 ]] && printf "%d%s" "$elapsedDays" "d:"
[[ "$elapsedHours" -gt 0 ]] && printf "%02d%s" "$elapsedHours" ":"
[[ "$elapsedMinutes" -gt 0 ]] && printf "%02d%s" "$elapsedMinutes" ":"
if [[ "$elapsedTime" -gt 59 ]]; then
	printf "%02d" "$elapsedSeconds"									#	At least 1 minute was already displayed, using colon delimited format.
else
	printf "%d%s" "$elapsedSeconds" " second"						#	Less than 1 minute, just display seconds.
	[[ "$elapsedSeconds" -ne 1 ]] && printf "s"						#	Handle the plural grammatically correct.
fi
printf ".\n\n"
#
#
######################################################################
#
#	The End
#
######################################################################
